unit CGI;

interface

uses
	Classes,
	DBTables,
	Forms,
	IniFiles,
	Messages,
	SysUtils,
	WinProcs,
	WinTypes;

type
	TCGIProfile = record
		AcceptTypes: TStringList;
		AuthType: String;
		AuthUser: String;
		ContentFile: String;
		ContentLength: LongInt;
		ContentType: String;
		DebugMode: ByteBool;
		ExecutablePath: String;
		ExtraHeaders: TStringList;
    GMTOffset: LongInt;
		LogicalPath: String;
		OutputFile: String;
		PhysicalPath: String;
    ProfileFile: String;
		QueryString: String;
		RemoteAddr: String;
		RemoteHost: String;
		RequestMethod: String;
		RequestProtocol: String;
		ServerAdmin: String;
		ServerName: String;
		ServerPort: Integer;
		ServerSoftware: String;
		TAPUser: String;
		Version: String;
	end;

	{ Define enumerated request methods for use in case statements }
	TRequestMethod = (rmGet, rmPost, rmTextSearch, rmHead, rmLink, rmUnlink, rmPut, rmOther);

	TCGI = class(TComponent)
		private
      FContentType: String;
     	FCookies: TStringList;
			FExternalTuples: TStringList;
			FFormTuples: TStringList;
			FHugeTuples: TStringList;
			FProfile: TCGIProfile;
			FStdOut: TMemoryStream;
      FResponseHeaders: TStringList;
			IniFile: TIniFile;

			procedure ErrorProc(Sender: TObject);
      procedure Initialize;
			function TranslateMethod: TRequestMethod;
      procedure ProcessMessages;
		public
			{ Methods }
			constructor Create(AOwner: TComponent); override;
			destructor Destroy; override;
			procedure Send(TextOut: String);
      procedure Include(Filename: String);
      procedure SendContent;

      function GetExternalSize(const Key: String): Integer;
      function GetExternalData(const Key: String; var Buffer: PChar): Integer;
			procedure ParseKeys(List: TStringList; const Params: String; const Delimit: Char);

			{ Run-Time Properties }
			property Cookies: TStringList read FCookies;
			property ExternalFields: TStringList read FExternalTuples;
			property FormFields: TStringList read FFormTuples;
			property HugeFields: TStringList read FHugeTuples;
			property Method: TRequestMethod read TranslateMethod;
			property Profile: TCGIProfile read FProfile;

			property ResponseHeaders: TStringList read FResponseHeaders;
      property StdOut: TMemoryStream read FStdOut;
		published
			{ Design-Time Properties and Events }
      property ContentType: String read FContentType write FContentType;
	end;

implementation

constructor TCGI.Create(AOwner: TComponent);
begin
	inherited Create(AOwner);
	with FProfile do begin
		AcceptTypes := TStringList.Create;
		ExtraHeaders := TStringList.Create;
	end;
 	FCookies := TStringList.Create;
	FFormTuples := TStringList.Create;
	FExternalTuples := TStringList.Create;
	FHugeTuples := TStringList.Create;
  FResponseHeaders := TStringList.Create;
  FContentType := 'text/html';

   { If it's run-time, let's load up our data! }
   if not (csDesigning in ComponentState) then	Initialize;
end;

destructor TCGI.Destroy;
begin
	with FProfile do begin
		AcceptTypes.Free;
		ExtraHeaders.Free;
  end;
  FCookies.Free;
	FFormTuples.Free;
	FExternalTuples.Free;
	FHugeTuples.Free;
  FResponseHeaders.Free;
	if Assigned(FStdOut) then begin
		FStdOut.Free;
  end;
  inherited Destroy;
end;

procedure TCGI.Initialize;
begin
	with FProfile do begin
		ProfileFile := ParamStr(1);
		IniFile := TIniFile.Create(ProfileFile);

		{ Read CGI and System Information }
		with IniFile do begin
			ServerSoftware := ReadString('CGI','Server Software', '');
			ServerName := ReadString('CGI', 'Server Name', '');
			ServerPort := ReadInteger('CGI', 'Server Port', -1);
			RequestProtocol := ReadString('CGI', 'Request Protocol', '');
			ServerAdmin := ReadString('CGI', 'Server Admin', '');
			Version := ReadString('CGI', 'CGI Version', '');
			RequestMethod := ReadString('CGI', 'Request Method', '');
			LogicalPath := ReadString('CGI', 'Logical Path', '');
			PhysicalPath := ReadString('CGI', 'Physical Path', '');
			ExecutablePath := ReadString('CGI', 'Executable Path', '');
			QueryString := ReadString('CGI', 'Query String', '');
			RemoteHost := ReadString('CGI', 'Remote Host', '');
			RemoteAddr := ReadString('CGI', 'Remote Address', '');
			AuthUser := ReadString('CGI', 'Authenticated User', '');
			TAPUser := ReadString('CGI', 'RFC-931 Identity', '');
			AuthType := ReadString('CGI', 'Authentication Method', '');
			ContentFile := ReadString('System', 'Content File', '');
			ContentType := ReadString('CGI', 'Content Type', '');
			ContentLength := ReadInteger('CGI', 'Content Length', 0);
      GMTOffset := ReadInteger('System', 'GMT Offset', -1);
			DebugMode := (ReadString('System', 'Debug Mode', 'No') = 'Yes');
			OutputFile := ExpandFileName(ReadString('System', 'Output File', ''));
		end;

		{ Open Output File; Get Accept Types and Extra Headers }
		IniFile.ReadSectionValues('Accept', AcceptTypes);
		IniFile.ReadSectionValues('Extra Headers', ExtraHeaders);
    if ExtraHeaders.Values['Cookie'] <> '' then
     	ParseKeys(FCookies,ExtraHeaders.Values['Cookie'],';');
		FStdOut := TMemoryStream.Create;
	end;

	{ Get Form Data }
	IniFile.ReadSectionValues('Form Literal', FFormTuples);
	IniFile.ReadSectionValues('Form External', FExternalTuples);
	IniFile.ReadSectionValues('Form Huge', FHugeTuples);
	IniFile.Free;

	{ Cycle Windows Messages -- Important! It lets the server know we're alive! }
	if Owner = nil then ProcessMessages else Application.ProcessMessages;
end;

procedure TCGI.Send(TextOut: String);
begin
	FStdOut.Write(TextOut[1], Length(TextOut));
end;

procedure TCGI.Include(Filename: String);
var
	IncludeFile: TFileStream;
begin
	IncludeFile := TFileStream.Create(Filename,fmOpenRead);
  try
    	StdOut.CopyFrom(IncludeFile,IncludeFile.Size);
  finally
  	IncludeFile.Free;
  end;
end;

procedure TCGI.SendContent;
var
	StdOutFile: TFileStream;
  TextOut: String;
  i: Integer;
  FM: word;
begin
	if FileExists(FProfile.OutputFile) then
  	FM :=fmOpenWrite or fmShareDenyNone
  else
  	FM := fmCreate;
	StdOutFile := TFileStream.Create(FProfile.OutputFile, FM);

	with ResponseHeaders do begin
  	Insert(0,'Content-Type: '+FContentType);
		Insert(0,'Content-Length: '+IntToStr(FStdOut.Size));
    Add(''); { Blank line to separate headers and data }
		for i := 0 to Count - 1 do begin
    	TextOut := Strings[i]+#13#10;
			StdOutFile.Write(TextOut[1], Length(TextOut));
    end;
  end;

  FStdOut.SaveToStream(StdOutFile);
  StdOutFile.Free;
end;

procedure TCGI.ErrorProc(Sender: TObject);
begin
	with FProfile do begin
		FStdOut.Seek(0,0);
		Send('<HTML>');
		Send('<HEAD>');
		Send('<TITLE>Error in ' + ExecutablePath + '</TITLE>');
		Send('<H1>Error in ' + ExecutablePath + '</H1>');
		Send('</HEAD>');
		Send('<BODY>');
		Send('An internal error has occurred in ' + ExecutablePath + '.<P>');
		Send('<I>Please</I> note what you were doing when this problem occurred,');
		Send('so we can identify and correct it. Write down the Web page you were using,');
		Send('any data you may have entered into a form or search box, and');
		Send('anything else that may help us duplicate the problem. Then contact the');
		Send('administrator of this service: ');
		Send('<A HREF="mailto:' + ServerAdmin + '">');
		Send('<ADDRESS>&lt;' + ServerAdmin + '&gt;</ADDRESS>');
		Send('</A></BODY></HTML>');
       SendContent;
       Halt;
	end;
end;

function TCGI.TranslateMethod: TRequestMethod;
const
	RequestMethods: array[Low(TRequestMethod)..High(TRequestMethod)] of String =
   			('GET','POST','TEXTSEARCH','HEAD','LINK','UNLINK','PUT','OTHER');
var
	i: TRequestMethod;
begin
	Result := High(TRequestMethod);
  	i := Low(TRequestMethod);
	with FProfile do
		while i < High(TRequestMethod) do begin
			if UpperCase(RequestMethod) = RequestMethods[i] then Result := i;
           Inc(i);
       end;
end;

function TCGI.GetExternalSize(const Key: String): Integer;
var
	i: Integer;
begin
	with FExternalTuples do begin
		i := Pos(' ',Values[Key]);
   	Result := StrToInt(Copy(Values[Key],i,Length(Values[Key])-(i-1)));
   end;
end;

function TCGI.GetExternalData(const Key: String; var Buffer: PChar): Integer;
var
	ExtFile, Filename: String;
	i, Size: Integer;
   FileStream: TFileStream;
begin
	with FExternalTuples do begin
		ExtFile := Values[Key];
		i := Pos(' ',ExtFile);
	   	Filename := Copy(ExtFile,1,i);
	   	System.Delete(ExtFile,1,i);
		Size := StrToInt(ExtFile);

   	FileStream := TFileStream.Create(Filename,fmOpenRead);
   	if StrBufSize(Buffer) >= Size then
   		Result := FileStream.Read(Buffer^,Size)
   	else
   		Result := 0;
   	FileStream.Destroy;
   end;
end;

procedure TCGI.ProcessMessages;
var
	Msg: TMsg;
begin
	while PeekMessage(Msg, 0, 0, 0, PM_REMOVE) do
   	if Msg.Message <> WM_QUIT then begin
			TranslateMessage(Msg);
			DispatchMessage(Msg);
       end else
       	Halt;
end;

procedure TCGI.ParseKeys(List: TStringList; const Params: String; const Delimit: Char);
var
	i: Integer;
  Pair: String;
begin
	Pair := '';
	for i := 1 to Length(Params) do
  	if Params[i] = Delimit then begin
     	if Length(Pair) > 0 then begin
        	List.Add(Pair);
        	Pair := '';
       end;
   end else
   		if ((Params[i] in ['_',' ']) and (Pair <> '')) then
        	Pair := Pair + ' '
     	else
        	Pair := Pair + Params[i];

	if Length(Pair) > 0 then List.Add(Pair);
end;

end.
